/* Part of the culibs project, <http://www.eideticdew.org/culibs/>.
 * Copyright (C) 2005--2007  Petter Urkedal <urkedal@nbi.dk>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <cu/idr.h>
#include <cu/util.h>
#include <cuoo/halloc.h>
#include <cuoo/intf.h>

cu_idr_t
cu_idr_by_cstr(char const *cstr)
{
    size_t cstr_size = strlen(cstr) + 1;
    size_t key_size = CUOO_HCOBJ_KEY_SIZE(sizeof(struct cu_idr_s) + cstr_size);
    cu_idr_t idr = cu_salloc(key_size + CUOO_HCOBJ_SHIFT);
    char *a_cstr = (char *)(idr + 1);
    idr->key_size = key_size;
    memcpy(a_cstr, cstr, cstr_size);
    memset(a_cstr + cstr_size, 0,
	   key_size - sizeof(struct cu_idr_s) - cstr_size + CUOO_HCOBJ_SHIFT);
    idr = cuoo_halloc(cu_idr_type(), key_size, (char *)idr + CUOO_HCOBJ_SHIFT);
    return idr;
}

cu_idr_t
cu_idr_by_charr(char const *arr, size_t charr_size)
{
    size_t cstr_size = charr_size + 1;
    size_t key_size = CUOO_HCOBJ_KEY_SIZE(sizeof(struct cu_idr_s) + cstr_size);
    cu_idr_t idr = cu_salloc(key_size + CUOO_HCOBJ_SHIFT);
    char *a_arr = (char *)(idr + 1);
    idr->key_size = key_size;
    memcpy(a_arr, arr, charr_size);
    memset(a_arr + charr_size, 0,
	   key_size - sizeof(struct cu_idr_s) - charr_size + CUOO_HCOBJ_SHIFT);
    idr = cuoo_halloc(cu_idr_type(), key_size, (char *)idr + CUOO_HCOBJ_SHIFT);
    return idr;
}

int
cu_idr_strcmp(cu_idr_t idr0, cu_idr_t idr1)
{
    return strcmp(cu_idr_to_cstr(idr0), cu_idr_to_cstr(idr1));
}

cu_clop_edef(cu_idr_strcmp_clop, int, cu_idr_t idr0, cu_idr_t idr1)
{
    return strcmp(cu_idr_to_cstr(idr0), cu_idr_to_cstr(idr1));
}

cu_clop_def(idr_key_size_clop, size_t, void *idr)
{
    return ((cu_idr_t)idr)->key_size;
}

cuoo_stdtype_t cuP_idr_type;

static void
idr_print(void *idr, FILE *out)
{
    char const *s = cu_idr_to_cstr(idr);
    fputs(s, out);
}

static cu_word_t
idr_impl(cu_word_t intf_number, ...)
{
    switch (intf_number) {
	case CUOO_INTF_PRINT_FN: return (cu_word_t)idr_print;
	default: return CUOO_IMPL_NONE;
    }
}

void
cuP_idr_init(void)
{
    cuP_idr_type = cuoo_stdtype_new_hcv(idr_impl, idr_key_size_clop);
}
